/*
   Basic software to run the Lexus GS450H hybrid transmission and inverter using the open source V1 controller

   Take an analog throttle signal and converts to a torque command to MG1 and MG2
   Feedback provided over USB serial
   V3 - Reverse added
   V4 - CAN specific to BMW E65 735i project
   V5 - WiFi connection on USART2 at 19200 baud
   V6 - add ISA CAN shunt connectivity. Note V2 hardware only.
   V7 - add HV precharge and control- oil pump relay=midpack and precharge contactor, out1= main contactor.

   Copyright 2020 T.Darby, D.Maguire, A.Kustusch
   openinverter.org
   evbmw.com

   TODO: 
   1) When in Drve (P) chrging plug is connected VCU does not switch into Charging mode
   2) When key on during charging, unplugging does not switch VCU into Drive mode
*/

#include <InterpolationLib.h>
#include <Metro.h>
#include <due_can.h>  //https://github.com/collin80/due_can
#include <due_wire.h> //https://github.com/collin80/due_wire
#include <DueTimer.h>  //https://github.com/collin80/DueTimer
#include <Wire_EEPROM.h> //https://github.com/collin80/Wire_EEPROM
#include <ISA.h>  //isa can shunt library

const int numValues = 10;
double xRegenRPM[10] = {0,100,1000,2000,3000,4000,5000,6000,8000,10000};
double yRegenTRQ[10] = {0,  0,  40,  80, 100, 100, 100, 100,  80,   60};

#define MG2MAXSPEED 10000
#define pin_inv_req 22

// Powertrain Modes
#define PARK 0
#define REVERSE 1
#define NEUTRAL 2
#define DRIVE 3

#define HVPreChargeRelay  33  //Orginally: OilPumpPower
#define OilPumpPWM  2
#define InvPower    34
#define Out1  50
#define TransSL1  47
#define TransSL2  44
#define TransSP   45

#define IN1   6  // Low[0] -> "Charging Mode"
#define IN2   7  // High[1] -> "Key ON"
#define Brake_In   62

#define TransPB1    40
#define TransPB2    43
#define TransPB3    42

#define HVPreset    300 //DC Bus Voltage at which to enable main contactor, my inverter resets value down to ~100V after motors reach ~4000RPM)

#define OilpumpTemp A7
#define TransTemp A4
#define MG1Temp A5
#define MG2Temp A6


//////////////// Constants ///////////////////////////////////////////////////////////////////
// Vehicle Operation Modes
const byte PROGRAMMING = 0;
const byte CHARGING = 1;
const byte DRIVING = 2;
const byte MONITORING = 3;

//////////////// Enums ///////////////////////////////////////////////////////////////////
typedef enum { NONE, DETAIL, CSV} debugLevels;

/*
enum VehicleMode: byte
{
   PROGRAMMING = 0,
   CHARGING = 1,
   DRIVING = 2,
   MONITORING = 3
};
VehicleMode VMode = VehicleMode::PROGRAMMING;
*/

//////////////// Global Variables ////////////////////////////////////////////////////////////
unsigned int _millis = 0;
debugLevels _debug = NONE;
bool _csv_log_started = false;
char* vehicleModeLabel = "";
int vehicleMode;
char charBuf[50];
word RPM;
byte Gcount; //gear display counter byte
unsigned int GLeaver;  //unsigned int to contain result of message 0x192. Gear selector lever position
int shiftPos; //contains byte to display gear position on dash
byte gear;
bool SportMode = false;
bool CurrentGear = HIGH;
bool RequestedGear = HIGH;
long GearShiftInit = 0;
byte mthCnt;
bool debug_console = false;
unsigned int line_no = 0;
unsigned int VCU_STATUS_DATA_LOW = 0;
unsigned int VCU_STATUS_DATA_HIGH = 0;
int MSG_DATA = 0;
byte inputValue = 0;
short BrakeForce = 0;    // Calculated brake force for regen
short BrakePressure = 4; // Brake system pressure from sensor (0-10)
short EngineBraking = 1; // Max engine braking torque (0-10)

//Regen Ramp
float RegenPercentage = 0;  // current percentage of regen (0-100)
long RampDuration = 500;   // ramp time [ms]
long RampStartTime = 0;    // start of current ramp

////////////////////////////////////////////////////////////////////////////////////////////////////
#define MAX_CAN_FRAME_DATA_LEN  8     // CAN frame max data length
#define CAN_VCU_STATUS_ID       0x07
#define CAN_BCM_STATUS_ID       0x08
CAN_FRAME VCU_STATUS_OUT_FRAME; //VCU Status CAN out frame
CAN_FRAME outframe;  //A structured variable according to due_can library for transmitting CAN data.

Metro timer_htm = Metro(10);
Metro timer_diag = Metro(1200); //lowering below 1100 ms causes WiFi Dashboard not to refresh
Metro timer_1ms = Metro(1);
Metro timer_10ms = Metro(10);
Metro timer_100ms = Metro(100);
Metro timer_250ms = Metro(250);
Metro timer_Frames100 = Metro(100);
Metro timer_Frames200 = Metro(200);
Metro timer_Frames4 = Metro(4); //ECM Frames

byte mth_data[100];
byte htm_data_setup[80] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 25, 0, 0, 0, 0, 0, 0, 0, 128, 0, 0, 0, 128, 0, 0, 0, 37, 1};
byte htm_data[80] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0, 0, 0, 0, 0};

unsigned short htm_checksum = 0,
               mth_checksum = 0,
               since_last_packet = 4000;

unsigned long  last_packet = 0;

volatile byte mth_byte = 0;

float dc_bus_voltage = 0, temp_inv_water = 0, temp_inv_inductor = 0; //just used for diagnostic output

short mg1_torque = 0,
      mg2_torque = 0,
      mg1_speed = -1,
      mg2_speed = -1;

short regen_torque = 0;
      
byte inv_status = 1;

bool htm_sent = 0,
     mth_good = 0;

int oil_power = 65;           //oil pump pwm value, 125 -> 50% pwm to oil pump at 1khz
int VacuumPressure = 0;       //value read from throttle input2 analog input
int Loop = 0;

int Throt1Pin = A0;           //throttle pedal analog inputs
int Throt2Pin = A1;

short TorqueReqest = 0;       //requested torque
short TorquePercent = 100;    //percentage of torque available, used to reduce torque for gear shifting and traction control
short TorqueAvailable = 0;    //available torque
bool TorqueRestoreOK = true;  //flag to restore full torque

unsigned long _preChargeBegin = 0;
unsigned long _preChargeEnd = 0;
unsigned long _preChargeAbort = 0;

bool T15Status;               //flag to keep status of Terminal 15 from CAS via CAN.
bool Key_ON;                  //flag for dash on can command.
bool can_status;              //flag for turning off and on can sending.
bool HV_Relays_OK;            //flag to indicate HV relays status
bool HV_Enabled;              //flag to indicate HV is ON
bool DC_Enabled;              //flag to indicate DC-DC is ON
bool Charge_Mode;             //flag to indicate Charging
bool BrakeDepressed;          //flsg to indicate Brake Pedal is pressed down

// detect Charge mode
bool get_chargeMode()
{
  bool chargeMode = false;
  if (!digitalRead(IN1))
  {
    chargeMode = true;
    VCU_STATUS_OUT_FRAME.data.byte[1] = 1;
  }
  else 
  {
    chargeMode = false;
    VCU_STATUS_OUT_FRAME.data.byte[1] = 0;
  }
  return chargeMode;
}

// detect operational mode (Charge vs Drive)
bool get_keyOn()
{
  bool keyOn = false;
  if (digitalRead(IN2)){
    keyOn = true;
    //TODO: figure out VCU Status Bit
    //VCU_STATUS_OUT_FRAME.data.byte[?] = 1;
  }
  return keyOn;
}

// this one is using 2 digital input pins
byte get_gear2()
{
  if (digitalRead(IN1) && !digitalRead(IN2)) { return (DRIVE); }
  else if (!digitalRead(IN1) && digitalRead(IN2)) { return (REVERSE); }
  else { return (NEUTRAL); }
}

// based on voltage of analog pin
byte get_gear()
{
  byte selectedGear = PARK;
  
  //accelerator input 2 mapping to gear selector
  int GearSelectorValue = analogRead(Throt2Pin); //50 to 370
  
  if (GearSelectorValue <= 275) selectedGear = PARK;
  else if (GearSelectorValue > 275 && GearSelectorValue <= 475) selectedGear = REVERSE;
  else if (GearSelectorValue > 475 && GearSelectorValue <= 675) selectedGear = NEUTRAL;
  else if (GearSelectorValue > 675) selectedGear = DRIVE;

  VCU_STATUS_OUT_FRAME.data.byte[2] = selectedGear;
  
  return selectedGear;
}

void setGear()
{
  if (RequestedGear ^ CurrentGear)    //gear shift requested
  {
    TorqueRestoreOK = false;          //put torque restore on hold
    TorquePercent = 0;                //reduce torque
    if (GearShiftInit = 0)
    {
      GearShiftInit = millis();
    }
    
    if(RequestedGear)                 //upshift into "High Gear"
    {
                                      //TODO: confirm torque reduction reading from MTH data
      //if (since_last_packet > 1000)   //wait for next 2 packages to be send to Inverter (1 every 1000us)
      if (millis() - GearShiftInit > 10)//wait 10ms
      {
        digitalWrite(TransSL1,LOW);
        digitalWrite(TransSL2,LOW);
      }
    
      //if(!digitalRead(TransPB1) ||    //when pressure is present on Brake 2
      //  abs(mg2_speed) < 100)         //or low RPM
      if (millis() - GearShiftInit > 20)//wait 10ms
      {
        CurrentGear = RequestedGear;  //update current 
        TorqueRestoreOK = true;       //allow to restore torque
        GearShiftInit = 0;
      }
    }
    else                              //downshift into "Low Gear"
    {
      if (millis() - GearShiftInit > 10)
      {
        digitalWrite(TransSL1,HIGH);
        digitalWrite(TransSL2,HIGH);
      }

      //if(!digitalRead(TransPB2) ||    //when pressure is present on Brake 2
      //  abs(mg2_speed) < 100)         //or low RPM
      if (millis() - GearShiftInit > 20)//wait 10ms
      {
        CurrentGear = RequestedGear;  //update current 
        TorqueRestoreOK = true;       //allow to restore torque
        GearShiftInit = 0;
      }
    }
  }
}

bool getBrake()
{
  if (!digitalRead(Brake_In))
  {
    //when brake pedal is just pressed, reset regen ramp
    if (!BrakeDepressed)
    {
      ResetRgenRamp(true);
    }
    //now update brke status
    BrakeDepressed = true;
    BrakeForce = BrakePressure;
  }
  else
  {
    BrakeDepressed = false;
    BrakeForce = 0;
  }
  return BrakeDepressed;
}

short get_torque()
{
  //accelerator pedal mapping to torque values here
  short throtVal = analogRead(Throt1Pin); // 20-607 (actual)
  if (throtVal < 45) throtVal = 40; //dead zone at start of throttle travel

  short torqueReq = 0;
  if (gear == DRIVE) 
  {
    // standard linear throttle
    //throtVal = map(throtVal, 40, 600, 0, 1500);

    // smooth full power
    if (throtVal <= 100)                          torqueReq = map(throtVal, 40, 100, 0, 160);
    else if(throtVal > 100 && throtVal <= 200)    torqueReq = map(throtVal, 101, 200, 161, 500);
    else if(throtVal > 200)                       torqueReq = map(throtVal, 201, 600, 501, 3500);    //max  3500
  }
  if (gear == REVERSE) torqueReq = map(throtVal, 40, 600, 0, -2500); //min -3500
  if (gear == PARK) torqueReq = 0; //no torque in park or neutral
  if (gear == NEUTRAL) torqueReq = 0; //no torque in park or neutral

  TorqueReqest = torqueReq;                              // requested torque
  TorqueAvailable = torqueReq * TorquePercent / 100;  // calculated available torque
  
  return TorqueAvailable;                             //return available torque
}

void ResetRgenRamp(bool brake)
{
  if (brake)
  {
    // for brake pedal
    RegenPercentage = 0;
  }
  else
  {
    // for engine braking
    RegenPercentage = 0;
  }
}

long RampUpRegen(double regen)
{
  if (RegenPercentage <= 0)
  {
    RampStartTime = millis();
    RegenPercentage = 0.1f;
  }
  else if (RegenPercentage < 1.0f)
  {
    long fromStart = millis() - RampStartTime;
    float perc = (float)fromStart / (float)RampDuration;
    if (perc > 1.0f)
    {
      RegenPercentage = 1.0f;
    }
    else
    {
      RegenPercentage = perc;
    }
  }
  else
  {
    RegenPercentage = 1.0f;
  }
  return RegenPercentage * (float)regen;
}

short CalcTorque( short throttle, short brake, short motorSpeed) //, short dir_selected)
{
  short torque;

  if(brake > 0)
  {
    if(motorSpeed > 100) //moving forward
    { 
      
      torque = (-1) * RampUpRegen(brake * Interpolation::Linear(xRegenRPM, yRegenTRQ, numValues, motorSpeed, true));
    }
    else
    {
      torque = 0;
    }
  }
  else if((throttle > 0) || (throttle < 0))
  {
    //reset regen ramp
    ResetRgenRamp(false);
    //let the throttle set the torque directly
    torque = throttle;
  }
  else //no throttle or brake applied so see if we need a negative torque
  {   
    if(motorSpeed > 100) //moving forward
    {
      //minimum torque to simulate engine braking
      torque = (-1) * RampUpRegen(EngineBraking * Interpolation::Linear(xRegenRPM, yRegenTRQ, numValues, motorSpeed, true));
    }
    else //not moving
    { 
      torque = 0;
    } 
  }
  return torque;
}

//Dilbert's regen implementation for ZombieVerter
//https://openinverter.org/forum/viewtopic.php?f=3&t=1277&p=23702&hilit=regen#p23702
short CalcTorque2( short throttle, short brake, short motor_speed) //, short dir_selected)
{
  short torque;

  if(brake > 0)
  {
    if(motor_speed > 1000) //moving forward
    { 
      torque = brake * (-1);//map(brake,0,100,0,-1000); //need to make this variable or atleast have a number of steps
    }
    else if ((motor_speed > 100) && (motor_speed <= 1000)) //moving forward slowly
    { 
      torque = map(motor_speed,100,1000,0,brake) * (-1); //ramp down
    }/*
    else if(motor_speed < -100) //moving back    
    {
      torque = map(brake,0,100,0,+100); 
    }*/
    else
    {
      torque = 0;
    }
  }
  else if((throttle > 0) || (throttle < 0))
  {
    //let the throttle set the torque directly
    torque = throttle;
  }
  else //no throttle or brake applied so see if we need a negative torque....maybe implement creep feature here, to up/down the torque to when in gear
  {   
    if(motor_speed > 1000) //moving forward
    {
      //minimum torque to simulate engine braking
      torque = EngineBraking * (-1);
    }
    if((motor_speed > 100) && (motor_speed <= 1000)) //moving forward slowly
    {
      //minimum torque to simulate engine braking
      torque = map(motor_speed,100,1000,0,EngineBraking) * (-1);
    }/*
    else if(motor_speed < -40) //going back
    {
      //minimum torque to simulate engine braking
      torque = 0;//200;
    }*/
    else//not moving
    { 
      torque = 0;
    } 
  }
  return torque;
}

ISA Sensor;  //Instantiate ISA Module Sensor object to measure current and voltage

#define SerialDEBUG SerialUSB

void setup() {

  //setup CAN0 communication////////////////
  Can0.begin(CAN_BPS_500K);  //Use for ISA shunt, comms, etc
  //TODO: Fix incoming communication from BCM
  //Can0.watchForRange(0x521, 0x528);  // listen for ISA messages(0x521-x528)
  //Can0.watchFor(CAN_BCM_STATUS_ID);  // only receive messages from BCM
  //Can0.attachCANInterrupt(IncomingCan0);

  // initialize VCU Status CAN frame object
  // Bytes: 0-HV, 1-Mode, 2-Gear
  VCU_STATUS_OUT_FRAME.id = CAN_VCU_STATUS_ID;
  VCU_STATUS_OUT_FRAME.length = MAX_CAN_FRAME_DATA_LEN;
  VCU_STATUS_OUT_FRAME.data.low = 0; //Set first four bytes (32 bits) all at once
  VCU_STATUS_OUT_FRAME.data.high = 0; //Set last four bytes (32 bits) all at once


  //setup CAN1 communication////////////////
  Can1.begin(CAN_BPS_500K);  //Use for communication with Jag's CAN Bus
  //Can1.watchForRange(0x130, 0x192);  // only receive messages from 0x130 to 0x192
  //Can1.attachCANInterrupt(Incoming);

  ///////////////////////////////////////////////////////////////////////////////////
  // WARNING - WARNING - WARNING - WARNING - WARNING - WARNING - WARNING - WARNING //
  ///////////////////////////////////////////////////////////////////////////////////
  // Below call will turn Digital Pin 50 (Out1) "High" until pin is configured     //
  // It may cause complications with HV PreCharge circuit since HV Main Contactor  //
  // may not open before PreCharge circuit closes.                                 //
  // Adding delay after resetting Pin 50 (Out1) should provide sufficient          //
  // workaround until issue is addressed at library level if even possible         //
  ///////////////////////////////////////////////////////////////////////////////////
  Sensor.begin(0, 500); //Start ISA object on CAN 0 at 500 kbps

  //set initial conditions/////////////////
  T15Status = false;
  Key_ON = false;
  can_status = false;
  RPM = 0;
  Gcount = 0x0d;
  gear = NEUTRAL;
  shiftPos = 0xb4; //select neutral
  //////////////////////////////////////////////

  pinMode(pin_inv_req, OUTPUT);
  digitalWrite(pin_inv_req, 1);
  pinMode(13, OUTPUT);  //led

  pinMode(HVPreChargeRelay, OUTPUT);  //Oil pump control relay is being used to control HV precharge contactor as not required seperate oil pump power control.
  digitalWrite(HVPreChargeRelay, LOW);  //turn off precharge

  pinMode(InvPower, OUTPUT);  //Inverter Relay
  pinMode(Out1, OUTPUT);  //GP output one
  delay(100); //Warning: this delay is necessary to allow time for Main HV Contactor to open before going into main program loop

  pinMode(TransSL1, OUTPUT); //Trans solenoids
  pinMode(TransSL2, OUTPUT); //Trans solenoids
  pinMode(TransSP, OUTPUT); //Trans solenoids

  digitalWrite(InvPower, LOW); //turn off at startup
  digitalWrite(Out1, LOW); //turn off HV main at startup
  digitalWrite(TransSL1, LOW); //turn off at startup
  digitalWrite(TransSL2, LOW); //turn off at startup
  digitalWrite(TransSP, LOW); //turn off at startup

  pinMode(IN1, INPUT); //Input 1
  pinMode(IN2, INPUT); //Input 2
  pinMode(Brake_In, INPUT); //Brake pedal input

  pinMode(TransPB1, INPUT); //Trans inputs
  pinMode(TransPB2, INPUT); //Trans inputs
  pinMode(TransPB3, INPUT); //Trans inputs
  
  Serial1.begin(250000);

  PIOA->PIO_ABSR |= 1 << 17;
  PIOA->PIO_PDR |= 1 << 17;
  USART0->US_MR |= 1 << 4 | 1 << 8 | 1 << 18;

  htm_data[63] = (-5000) & 0xFF; // regen ability of battery
  htm_data[64] = ((-5000) >> 8);

  htm_data[65] = (27500) & 0xFF; // discharge ability of battery
  htm_data[66] = ((27500) >> 8);

  Serial2.begin(19200); //setup serial 3 for wifi access

  SerialDEBUG.begin(115200);
}

// detect vehicle operation mode
byte get_VehicleMode()
{
  byte currentMode = PROGRAMMING;
 
  if (!Key_ON && !Charge_Mode)      //Programming
  {
    currentMode = PROGRAMMING;
    vehicleModeLabel = "PROGRAMMING";
  }
  else if (!Key_ON && Charge_Mode)  //Charging
  {
    currentMode = CHARGING;
    vehicleModeLabel = "CHARGING";
  }
  else if (Key_ON && !Charge_Mode)  //Driving
  {
    currentMode = DRIVING;
    vehicleModeLabel = "DRIVING";
  }
  else if (Key_ON && !Charge_Mode)  //Monitoring
  {
    currentMode = MONITORING;
    vehicleModeLabel = "MONITORING";
  }
 
  return currentMode;
}

void Check_DC_Bus_Voltage()
{
  // Test if DC Bus Voltage is within 10% of Battery Voltage
  // lower value indicates interruption like open contactor or other problem
  // This check is also necessary when VCU is powered through USB port to prevent damage of pre-charge circuit
  if (Sensor.Voltage2 < Sensor.Voltage1*0.9 && HV_Enabled)  
  {
    HV_Off();               // turn off relays and update HV Status
    HV_Relays_OK = false;   // force relay test before HV is enabled again
  }
}

bool Check_HV_Relays_OLD()
{
  if (!HV_Relays_OK) 
  {
    digitalWrite(HVPreChargeRelay, LOW);  //turn off HV Negative and PreCharge relays
    digitalWrite(Out1, LOW);              //turn off HV Positive relay
  
    for (int i = 0; i < 100; i++)  {
      // Test for "welded" or closed Positive HV relay
      // DC Bus Voltage must be much less then Battery Voltage
      //if (true) //TODO: for testing only
      if (Sensor.Voltage2 < Sensor.Voltage1*0.5)
      {
        // main HV relay is not stuck, it is safe to turn Negative HV and PreCharge relays
        HV_Relays_OK = true;
        break;
      }
      Loop++;
      delay(10); // wait for relay to open
    }
  }
  return HV_Relays_OK;
}

bool Check_HV_Relays()
{
  if (!HV_Relays_OK) 
  {
    digitalWrite(HVPreChargeRelay, LOW);  //turn off HV Negative and PreCharge relays
    digitalWrite(Out1, LOW);              //turn off HV Positive relay
  
    // Test for "welded" or closed Positive HV relay
    // DC Bus Voltage must be much less then Battery Voltage
    //if (true) //TODO: for testing only
    if (Sensor.Voltage2 < Sensor.Voltage1*0.5)
    {
      // main HV relay is not stuck, it is safe to turn Negative HV and PreCharge relays
      HV_Relays_OK = true;
    }
    Loop++;
  }
  return HV_Relays_OK;
}

void Control_HV()
{
  if (!HV_Enabled) Check_HV_Relays();

  // Enable HV if "HV Relays OK" and "Key ON" and "Not Charging"
  if (HV_Relays_OK && Key_ON && !Charge_Mode)
  {
    if (!HV_Enabled) HV_On(false);
  }
  // Sort of emergency mode, since relays are welded no point of driving them but simply update HV status which should be enabled at this point
  else if (Loop > 100 && !HV_Relays_OK && Key_ON && !Charge_Mode)
  {
    if (!HV_Enabled) HV_On(true);
  }
  else
  {
    HV_Off();
  }
}

void HV_On_Cool()
{
  if (_preChargeBegin == 0 && _preChargeAbort == 0) //start pre-charge cycle
  {
    _preChargeBegin = millis(); //record begin of pre-charge cycle
    _preChargeEnd = 0; //reset end timer
    digitalWrite(HVPreChargeRelay, HIGH);  //turn on hv precharge contactors
  }
  else if (_preChargeAbort > 0) //restart pre-charge after cooling down resistor
  {
    if (millis() - _preChargeAbort > 1000)
    {
      _preChargeAbort = 0;
      _preChargeBegin = millis(); //record begin of pre-charge cycle
      digitalWrite(HVPreChargeRelay, HIGH);  //turn on pre-charge contactors
    }
  }
  else //check lenght of pre-charge cycle and abort if necessary
  {
    // abort pre-charge cycle to prevent burn out of resistor
    if (millis() - _preChargeBegin > 3000)
    {
      _preChargeBegin = 0;
      _preChargeAbort = millis(); //record pre-charge abort
      digitalWrite(HVPreChargeRelay, LOW);  //turn of hv precharge and pack split contactors
    }
  }

  //complete pre-charge
  if (Sensor.Voltage2 > Sensor.Voltage1*0.9) //get DC Bus Voltage from ISA since Inverter is broken and does not report correct voltage.
  {
    _preChargeEnd = millis(); //record end of pre charge cycle
    //_preChargeAbort = 0; //reset abort time
    digitalWrite(Out1, HIGH); //turn on hv main once precharge complete
    HV_Enabled = true;
    VCU_STATUS_OUT_FRAME.data.byte[0] = 1;
  }
}

void HV_On(bool weldedRelays)
{
  if (_preChargeBegin == 0) //start pre-charge cycle
  {
    _preChargeBegin = millis(); //record begin of pre-charge cycle
    _preChargeEnd = 0; //reset end timer
    if (!weldedRelays) digitalWrite(HVPreChargeRelay, HIGH);  //turn on hv precharge contactors
  }
  else //check lenght of pre-charge cycle and abort if necessary
  {
    // abort pre-charge cycle to prevent burn out of resistor (usually precharge takes under 300ms)
    if (millis() - _preChargeBegin > 20000 && _preChargeAbort == 0)
    {
      _preChargeAbort = millis(); //record pre-charge abort
      _preChargeEnd = 0;
      if (!weldedRelays) digitalWrite(HVPreChargeRelay, LOW);  //turn of hv precharge and pack split contactors
    }
  }
  
  //complete pre-charge
  if ((_preChargeAbort ==0) && (Sensor.Voltage2 > Sensor.Voltage1*0.9)) //get DC Bus Voltage from ISA since Inverter is broken and does not report correct voltage.
  {
    _preChargeEnd = millis(); //record end of pre charge cycle
    if (!weldedRelays) digitalWrite(Out1, HIGH); //turn on hv main once precharge complete
    HV_Enabled = true;
    VCU_STATUS_OUT_FRAME.data.byte[0] = 1;
  }
}

void HV_Off()
{
  _preChargeBegin = 0;
  _preChargeEnd = 0;
  _preChargeAbort = 0;
  digitalWrite(Out1, LOW); //turn of hv main contactor
  digitalWrite(HVPreChargeRelay,LOW);  //turn of hv precharge and pack split contactors
  HV_Enabled = false;
  VCU_STATUS_OUT_FRAME.data.byte[0] = 0;
}


void handle_wifi() {
  /*
    Routine to send data to wifi on serial 2
    The information will be provided over serial to the esp8266 at 19200 baud 8n1 in the form :
    vxxx,ixxx,pxxx,mxxxx,nxxxx,oxxx,rxxx,qxxx* where :

    v=pack voltage (0-700Volts)
    i=current (0-1000Amps)
    p=power (0-300kw)
    m=mg1 rpm (0-10000rpm)
    n=mg2 rpm (0-10000rpm)
    o=mg1 temp (-20 to 120C)
    r=mg2 temp (-20 to 120C)
    q=oil pressure (0-100%)
    =end of string
    xxx=three digit integer for each parameter eg p100 = 100kw.
    updates will be every 100ms approx.

    v100,i200,p35,m3000,n4000,o20,r100,q50*
  */

  //Serial2.print("v100,i200,p35,m3000,n4000,o20,r30,q50*"); //test string
  //Serial2.print("v111,i222,p333,m444,n555,o666,r777,q888*"); //test string
  //Serial2.print("v355,i50,p17.75,m14.8,n1000,o32,r25,q40*"); //test string

  bool test = false;
  if (test)
  {
    // random test values
    long  test_pack_voltage = random(300,400);
    long  test_current = random(1000);
    long  test_power = (test_pack_voltage * test_current) / 1000;
    long  test_kilowat_hours = random(0,14.8);
    long  test_mg2_speed = random(10000);
    long  test_temperature1 = random(-40,120);
    long  test_temperature2 = random(-40,120);
    long  test_pressure = random(100);
  
    Serial2.print("v");   //Pack Voltage
    Serial2.print(test_pack_voltage);
    Serial2.print(",i");  //Current
    Serial2.print(test_current);    //Current derived from ISA shunt
    Serial2.print(",p");  //Power
    Serial2.print(test_power);         //Power derived from ISA shunt
    Serial2.print(",m");  //Kilowatt Hours
    Serial2.print(test_kilowat_hours);
    Serial2.print(",n");  //Motor RPM
    Serial2.print(test_mg2_speed);
    Serial2.print(",o");  //Motor Temperature
    Serial2.print(test_temperature1);  //Value derived from ISA shunt, +10000 to correct parsing issue on the WEB which strps 1st digit
    Serial2.print(",r");  //Coolant Temperature
    Serial2.print(test_temperature2);
    Serial2.print(",q");  //Oil Pressure
    Serial2.print(test_pressure);// change oil pump pwm value to percent.
    Serial2.print("*");// end of data indicator
  }
  else
  {
    Serial2.print("v");   //Pack Voltage
    Serial2.print(Sensor.Voltage1);   //Voltage derived from ISA Shunt
    Serial2.print(",i");  //Current
    Serial2.print(Sensor.Amperes);    //Current derived from ISA shunt
    Serial2.print(",p");  //Power
    Serial2.print(Sensor.KW);         //Power derived from ISA shunt
    Serial2.print(",m");  //Kilowatt Hours
    Serial2.print(Sensor.kwh / 1000.0f); //Value derived from ISA shunt
    Serial2.print(",n");  //Motor RPM
    Serial2.print(Abs(mg2_speed));
    Serial2.print(",o");  //Motor Temperature
    Serial2.print(temp_inv_inductor);  //Value derived from ISA shunt, +10000 to correct parsing issue on the WEB which strps 1st digit
    Serial2.print(",r");  //Coolant Temperature
    Serial2.print(temp_inv_water);
    Serial2.print(",q");  //Oil Pressure
    Serial2.print(map(oil_power, 0, 255, 0, 100));// change oil pump pwm value to percent.
    Serial2.print("*");// end of data indicator

    //SerialDEBUG.print("WIFI[Sensor.Voltage1]: "); SerialDEBUG.println(Sensor.Voltage1);
  }
}

void control_inverter() {

  int speedSum = 0;

  if (timer_htm.check()) //prepare htm data
  {
    if (mth_good)
    {
      dc_bus_voltage = (((mth_data[82] | mth_data[83] << 8) - 5) / 2);
      temp_inv_water = (mth_data[42] | mth_data[43] << 8);
      temp_inv_inductor = (mth_data[86] | mth_data[87] << 8);
      mg1_speed = mth_data[6] | mth_data[7] << 8;
      mg2_speed = mth_data[31] | mth_data[32] << 8;
    }
    gear = get_gear();
    //mg2_torque = get_torque(); // -3500 (reverse) to 3500 (forward)
    regen_torque = CalcTorque(get_torque(), BrakeForce, mg2_speed); //, short dir_selected)
    mg2_torque = regen_torque;
    
    mg1_torque = ((mg2_torque * 5) / 4);

    if ((mg2_speed > MG2MAXSPEED) || (mg2_speed < - MG2MAXSPEED))mg2_torque = 0;
    if (gear == REVERSE)mg1_torque = 0;

    // For testing only
    //if(gear==DRIVE)mg1_torque=0; //disable MG1 in DRIVE
    //if(gear==DRIVE)mg2_torque=0; //disable MG2 in DRIVE

    //speed feedback
    speedSum = mg2_speed + mg1_speed;
    speedSum /= 113;
    htm_data[0] = (byte)speedSum;
    htm_data[75] = (mg1_torque * 4) & 0xFF;
    htm_data[76] = ((mg1_torque * 4) >> 8);

    //mg1
    htm_data[5] = (mg1_torque * -1) & 0xFF; //negative is forward
    htm_data[6] = ((mg1_torque * -1) >> 8);
    htm_data[11] = htm_data[5];
    htm_data[12] = htm_data[6];

    //mg2
    htm_data[26] = (mg2_torque) & 0xFF; //positive is forward
    htm_data[27] = ((mg2_torque) >> 8);
    htm_data[32] = htm_data[26];
    htm_data[33] = htm_data[27];

    //checksum
    htm_checksum = 0;
    for (byte i = 0; i < 78; i++)htm_checksum += htm_data[i];
    htm_data[78] = htm_checksum & 0xFF;
    htm_data[79] = htm_checksum >> 8;
  }

  since_last_packet = micros() - last_packet;

  if (since_last_packet >= 4000) //read mth
  {
    htm_sent = 0;
    mth_byte = 0;
    mth_checksum = 0;

    for (int i = 0; i < 100; i++)mth_data[i] = 0;
    while (Serial1.available()) {
      mth_data[mth_byte] = Serial1.read();
      mth_byte++;
    }

    for (int i = 0; i < 98; i++)mth_checksum += mth_data[i];
    if (mth_checksum == (mth_data[98] | (mth_data[99] << 8)))mth_good = 1; else mth_good = 0;
    last_packet = micros();
    digitalWrite(pin_inv_req, 0);
  }

  since_last_packet = micros() - last_packet;

  if (since_last_packet >= 10)digitalWrite(pin_inv_req, 1);

  if (since_last_packet >= 1000)
  {
    if (!htm_sent && inv_status == 0) {
      for (int i = 0; i < 80; i++)Serial1.write(htm_data[i]);
    }
    else if (!htm_sent && inv_status != 0) {
      for (int i = 0; i < 80; i++)Serial1.write(htm_data_setup[i]);
      if (mth_data[1] != 0) inv_status--;
    }
    htm_sent = 1;
  }
}

void diag_mth()
{
  /*
  //mask just hides any MTH data byte which is represented here with a 0. Useful for debug/discovering.
  bool mth_mask[100] = {
    0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
    1, 1, 0, 0, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 0, 0, 1, 1, 1, 1, 0, 0, 1,
    1, 1, 0, 0, 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    1, 1, 0, 0, 1, 1, 0, 0, 1, 1,
    1, 1, 1, 1, 1, 1, 1, 1, 0, 0,
  };

  SerialDEBUG.print("\n");
  SerialDEBUG.println("\t0\t1\t2\t3\t4\t5\t6\t7\t8\t9");
  SerialDEBUG.println("   ------------------------------------------------------------------------------");
  for (int j = 0; j < 10; j++)
  {
    SerialDEBUG.print(j * 10); if (j == 0)SerialDEBUG.print("0"); SerialDEBUG.print(" |\t");
    for (int k = 0; k < 10; k++)
    {
      if (mth_mask[j * 10 + k])SerialDEBUG.print(mth_data[j * 10 + k]); else SerialDEBUG.print (" ");
      SerialDEBUG.print("\t");
    }
    SerialDEBUG.print("\n");
  }
  SerialDEBUG.print("\n");
  */

  serial_printf(SerialUSB, "=========================================\n");
  serial_printf(SerialUSB, "Millis:\t%ims\n", _millis);
  serial_printf(SerialUSB, "=========================================\n\n");

  //TODO: figure out why label is not always showing
  SerialDEBUG.print("Mode:\t\t\t");SerialDEBUG.print(vehicleModeLabel);SerialDEBUG.print("[");SerialDEBUG.print(vehicleMode);SerialDEBUG.println("]");
  //serial_printf(SerialUSB, "Mode:\t\t\t%s[%i]\n", vehicleModeLabel, vehicleMode);
  serial_printf(SerialUSB, "MTH Valid:\t\t%y\tChecksum: %i\n", mth_good, mth_checksum);
  serial_printf(SerialUSB, "HV Relays OK:\t\t%y\tLoop:     %i\n", HV_Relays_OK, Loop);
  serial_printf(SerialUSB, "HV Enabled:\t\t%y\n", HV_Enabled);

  // Pre-Charg Timing
  /*
  serial_printf(SerialUSB, "PreChargeBegin:\t\t%ims\n", _preChargeBegin);
  serial_printf(SerialUSB, "PreChargeEnd:\t\t%ims\n", _preChargeEnd);
  serial_printf(SerialUSB, "PreChargeAbort:\t\t%ims\n", _preChargeAbort);
  */
  
  // Pre-Charge Status
  if (_preChargeBegin > 0 && _preChargeEnd > 0 && _preChargeAbort == 0)
  {
    serial_printf(SerialUSB, "Pre-Charge completed:\t%ims\n", _preChargeEnd - _preChargeBegin);
  }
  else if (_preChargeBegin > 0 && _preChargeEnd == 0 && _preChargeAbort == 0)
  {
    serial_printf(SerialUSB, "Pre-Charge in progress:\t%ims\n", millis() - _preChargeBegin);
  }
  else if (_preChargeBegin > 0 && _preChargeEnd == 0 && _preChargeAbort > 0)
  {
    serial_printf(SerialUSB, "Pre-Charge aborted:\t%ims\n", _preChargeAbort - _preChargeBegin);
  }
  else
  {
    serial_printf(SerialUSB, "Pre-Charge wait:\t-\n");
  }
  
  serial_printf(SerialUSB, "DC Bus (Inverter):\t%fV\n", dc_bus_voltage);
  serial_printf(SerialUSB, "DC Bus (Shunt):\t\t%fV\n", Sensor.Voltage2);
  serial_printf(SerialUSB, "Gear:\t\t\t%i; Sport Mode[%y]\n", gear, SportMode);
  serial_printf(SerialUSB, "Gear Speed:\t\tCurrent[%h], Requested[%h]\n", CurrentGear, RequestedGear);
  serial_printf(SerialUSB, "Trans Pressure:\t\tPB1[%o], PB2[%o], PB3[%o]\n", !digitalRead(TransPB1), !digitalRead(TransPB2), !digitalRead(TransPB3));
  serial_printf(SerialUSB, "Trans Gear Solenoids:\tSL1[%o], SL2[%o], SL3[%o]\n", digitalRead(TransSL1), digitalRead(TransSL2), digitalRead(TransSP));
   
  serial_printf(SerialUSB, "Torque Reqest:\t\t%i\n", TorqueReqest);
  serial_printf(SerialUSB, "Torque Percent:\t\t%i\n", TorquePercent);
  serial_printf(SerialUSB, "Torque Available:\t%i\n", TorqueAvailable);
  serial_printf(SerialUSB, "Regen Torque:\t\t%i\n", regen_torque);
  serial_printf(SerialUSB, "Brake Pressure:\t\t%i\n", BrakePressure);
  serial_printf(SerialUSB, "Engine Braking:\t\t%i\n", EngineBraking);
  serial_printf(SerialUSB, "Ramp Duration:\t\t%i\n", RampDuration);
  //serial_printf(SerialUSB, "Torque Available:\tH[%i], L[%i]\n", (mg2_torque) & 0xFF, ((mg2_torque) >> 8));
  serial_printf(SerialUSB, "Oil Pump PWM:\t\t%i\n", oil_power);
  
  // Motors
  serial_printf(SerialUSB, "MG1 - Speed: %irpm\tPosition: %i\n", mg1_speed, mth_data[12] | mth_data[13] << 8);
  serial_printf(SerialUSB, "MG2 - Speed: %irpm\tPosition: %i\n", mg2_speed, mth_data[37] | mth_data[38] << 8);

  // Inverter Temperatures
  serial_printf(SerialUSB, "Water Temp:\t\t%1fC\n", temp_inv_water);
  serial_printf(SerialUSB, "Inductor Temp:\t\t%1fC\n", temp_inv_inductor);
  //serial_printf(SerialUSB, "Another1 Temp:\t\t%fC\n", mth_data[88] | mth_data[89] << 8);
  //serial_printf(SerialUSB, "Another2 Temp:\t\t%fC\n", mth_data[41] | mth_data[40] << 8);
  
  // ISA Sensor
  serial_printf(SerialUSB, "Battery Voltage:\tV1[%fV], V2[%fV], V3[%fV]\n", Sensor.Voltage1, Sensor.Voltage2, Sensor.Voltage3);
  serial_printf(SerialUSB, "Instant Current:\t%fA\n", Sensor.Amperes);
  serial_printf(SerialUSB, "Instant Power:\t\t%fKW\n", Sensor.KW);
  serial_printf(SerialUSB, "Ampere Hours:\t\t%3fkWh\n", Sensor.ah);
  serial_printf(SerialUSB, "Kilowatt Hours:\t\t%3fkWh\n\n", Sensor.kwh / 1000.0f);

  // Misc Values
  //serial_printf(SerialUSB, "Analog A2 Value:\t%i\n", analogRead(Throt2Pin));
  
  serial_printf(SerialUSB, "...\n");
}

void csv_info()
{
  if (!_csv_log_started) {
    _csv_log_started = true;

    //Header Line
    SerialDEBUG.print("Millis,");
    SerialDEBUG.print("MTH Valid,Checksum,DC Bus [V],Gear,Throttle,MG1 [RPM],MG2 [RPM],Water Temp [C],Inductor Temp [C],Temp 1 [C],Temp 2 [C],");
    SerialDEBUG.print("Battery Voltage [V],DC Bus Voltage [V],Current [A],Power [kW],Charge [Ah],Energy [kWh],");
    SerialDEBUG.print("Vacuum");
    SerialDEBUG.print("\n");
  }

  //CSV Data
  SerialDEBUG.print(_millis); SerialDEBUG.print(",");
  if (mth_good)SerialDEBUG.print("Yes"); else SerialDEBUG.print("No"); SerialDEBUG.print(",");
  SerialDEBUG.print(mth_checksum); SerialDEBUG.print(",");
  if (dc_bus_voltage >= 0) SerialDEBUG.print(dc_bus_voltage); else SerialDEBUG.print("----"); SerialDEBUG.print(",");
  SerialDEBUG.print(gear); SerialDEBUG.print(",");
  SerialDEBUG.print(mg2_torque); SerialDEBUG.print(",");
  SerialDEBUG.print(mg1_speed); SerialDEBUG.print(",");
  SerialDEBUG.print(mg2_speed); SerialDEBUG.print(",");
  SerialDEBUG.print(temp_inv_water); SerialDEBUG.print(",");
  SerialDEBUG.print(temp_inv_inductor); SerialDEBUG.print(",");
  SerialDEBUG.print(mth_data[88] | mth_data[89] << 8); SerialDEBUG.print(",");
  SerialDEBUG.print(mth_data[41] | mth_data[40] << 8); SerialDEBUG.print(",");
  SerialDEBUG.print(Sensor.Voltage1); SerialDEBUG.print(",");
  SerialDEBUG.print(Sensor.Voltage2); SerialDEBUG.print(",");
  SerialDEBUG.print(Sensor.Amperes); SerialDEBUG.print(",");
  SerialDEBUG.print(Sensor.KW); SerialDEBUG.print(",");
  SerialDEBUG.print(Sensor.ah); SerialDEBUG.print(",");
  SerialDEBUG.print(Sensor.kwh / 1000.0f); SerialDEBUG.print(",");
  SerialDEBUG.print(VacuumPressure); SerialDEBUG.print("\n");
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
///////Handle incomming pt can messages on CAN0
////////////////////////////////////////////////////////////////////////////////////////////////////
void IncomingCan0 (CAN_FRAME *frame) {

  ///////////Message from BCM indicating status of DC-DC converter

  // Process messages from BCM
  if (frame->id == CAN_BCM_STATUS_ID)
  {
    //Update HV Status
    if (frame->data.byte[0] > 0) DC_Enabled = true;
    else DC_Enabled = false;
    
    //Update Gear Ratio (Hi/Lo)
    setGearRatio(!frame->data.byte[0]);
  }
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
///////Handle incomming pt can messages from the car here
////////////////////////////////////////////////////////////////////////////////////////////////////
void Incoming (CAN_FRAME *frame) {

  ///////////Message from CAS on 0x130 byte one for Terminal 15 wakeup

  if (frame->id == 0x130)
  {
    if (frame->data.byte[0] == 0x45) T15Status = true; //if the cas sends 0x45 in byte 0 of id 0x130 we have a run command
    else T15Status = false;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  //////////////////////Decode gear selector , update inverter and display back onto cluster in car.
  if (frame->id == 0x192)
  {
    GLeaver = frame->data.low;
    GLeaver = GLeaver & 0x00ffffff; //mask off byte 3
    // Serial.println(GLeaver,HEX);
    switch (GLeaver) {
      case 0x80006a:  //not pressed
        break;
      case 0x80506a:  //park button pressed
        gear = PARK;
        shiftPos = 0xe1;
        break;
      case 0x800147:  //R position
        gear = REVERSE;
        shiftPos = 0xd2;
        break;
      case 0x80042d: //R+ position
        gear = NEUTRAL;
        shiftPos = 0xb4; //select Neutral on overpress
        break;
      case 0x800259:  //D pressed
        gear = DRIVE;
        shiftPos = 0x78;
        break;
      case 0x800374:  //D+ pressed
        gear = NEUTRAL;
        shiftPos = 0xb4; //select Neutral on overpress.
        break;
      case 0x81006a:  //Left Back button pressed
        break;
      case 0x82006a:  //Left Front button pressed
        break;
      case 0x84006a:  //right Back button pressed
        break;
      case 0x88006a:  //right Front button pressed
        break;
      case 0xa0006a:  //  S-M-D button pressed
        break;
      default:
      {
      }
    }
  }
}

/////////////////this can id must be sent once at T15 on to fire up the instrument cluster/////////////////////////
void DashOn() {

  outframe.id = 0x332;            // dash on message
  outframe.length = 2;            // Data payload 2 bytes
  outframe.extended = 0;          // standard id
  outframe.rtr = 1;               // No request
  outframe.data.bytes[0] = 0x61;  // sets max rpm on tach (temp thing)
  outframe.data.bytes[1] = 0x82;
  Can1.sendFrame(outframe);
}

/////////////Send frames every 4ms and send/rexeive inverter control serial data ///////////////////////////////////////

void Frames4MS()
{
  if (timer_Frames4.check())
  {
    if (can_status)
    {
      // ECM Status: Brake pedal, Engine RPM,  //////////////////////////////////////////////////////////////

      int ThrotPos = analogRead(Throt1Pin); // Actual values 20-607
      
      RPM = abs(mg2_speed);
      if (RPM < 10)
      {
        //RPM = 10; // 10 RPM is enough to prevent "Engine Stalled" warning (shows as 0 on Tach)
        RPM = 100;  // 100 RPM is necessary to engage "Heater Pump" (shows as 0 on Tach)
      }

      outframe.id = 0x12C;            // Set our transmission address ID
      outframe.length = 8;            // Data payload 8 bytes
      outframe.extended = 0;          // Extended addresses: 0=11-bit 1=29bit
      outframe.rtr = 1;               // No request
      outframe.data.bytes[0] = map(ThrotPos, 0, 650, 0, 100); // Estimated torque (1bit=1%)
      outframe.data.bytes[1] = map(ThrotPos, 0, 650, 0, 255); // Throttle position (0.4bit=1%)
      outframe.data.bytes[2] = outframe.data.bytes[1];        // Pedal position (0.4bit=1%)
      outframe.data.bytes[3] = 0x00;                          // Torque reduction acknowledge
      outframe.data.bytes[4] = highByte(RPM);                 // Engine RPM MSB
      outframe.data.bytes[5] = lowByte(RPM);                  // Engine RPM LSB
      outframe.data.bytes[6] = 0x00;                          // Brake pedal released 0x00
      outframe.data.bytes[7] = 0x00;                          // Unknown, always 0x00

      if (getBrake())
      {
        outframe.data.bytes[6] = 0x01;                        // Brake pedal pressed 0x01
      }

      Can1.sendFrame(outframe);
      ///////////////////////////////////////////////////////////////////////////////
    }
  }
}


void Frames100MS()
{
  if (timer_Frames100.check())
  {
    if (can_status)
    {
      // ECM Diagnostic: OBD MIL, Fault codes //////////////////////////////////////////////////////////////

      byte MIL_Code = 0x81; //MIL On
      if (HV_Enabled)
      {
        MIL_Code = 0x80; //MIL Off
      }
      
      outframe.id = 0x44C;            // Set our transmission address ID
      outframe.length = 6;            // Data payload 6 bytes
      outframe.extended = 0;          // Extended addresses: 0=11-bit 1=29bit
      outframe.rtr = 1;               // No request

      //int analog1 = analogRead(Throt1Pin);
      //int temp = map(analog1, 0, 765, 80, 125); // for testing low range
      //int temp = map(analog1, 0, 765, 150, 165);// for testing high range
      
      outframe.data.bytes[0] = temp_inv_water + 120;  //TODO:  +120 will work for temp 25C-up   //above 30°C (86°F) is necessary for "Heater Pump" to operate    // Coolant temp (80=0%, 100=12.5%, 110=25%, 118=40%, 125=50%, 150=50%, 150=50%, 155=50%, 160=95%, 165=100%)
      outframe.data.bytes[1] = MIL_Code;//Actual B1,F1,F3,F5  // OBD MIL, Throttle malfunction
      outframe.data.bytes[2] = 0x00;                          // Pedal position (0.4bit=1%)
      outframe.data.bytes[3] = 0x00;                          // Engine fault codes 
      outframe.data.bytes[4] = 0x00;                          // Engine RPM MSB
      outframe.data.bytes[5] = 0x00;                          // Engine RPM LSB

      Can1.sendFrame(outframe);
      ///////////////////////////////////////////////////////////////////////////////
    }
  }
}


////////////Send these frames every 200ms /////////////////////////////////////////
void Frames200MS()
{
  if (timer_Frames200.check())
  {
    digitalWrite(13, !digitalRead(13)); //blink led every time we fire this interrrupt.

    if (can_status)
    {
      // ECM heartbeet //////////////////////////////////////////////////////////////
      outframe.id = 0x7D0;            // current selected gear message
      outframe.length = 1;            // Data payload 1 byte
      outframe.extended = 0;          // Extended addresses - 0=11-bit 1=29bit
      outframe.rtr = 1;               // No request
      outframe.data.bytes[0] = 0xC4;

      Can1.sendFrame(outframe);
      ///////////////////////////////////////////////////////////////////////////////
      
      // TCM heartbeet //////////////////////////////////////////////////////////////
      outframe.id = 0x7D1;            // current selected gear message
      outframe.length = 1;            // Data payload 1 byte
      outframe.extended = 0;          // Extended addresses - 0=11-bit 1=29bit
      outframe.rtr = 1;               // No request
      outframe.data.bytes[0] = 0xC4;

      Can1.sendFrame(outframe);
      ///////////////////////////////////////////////////////////////////////////////
    }

      //VCU_STATUS_OUT_FRAME bytes
      // 0: HV (Off=0; On=1)
      // 1: Charge Mode (No=0; Yes=1)
      // 2: Gear (0, 1, 2, 3)
      // 3: Sensor.kwh/100
      // 4: MG2 RPM/100

      //Set MG2 RPM/100
      int mg2rpmby100 = mg2_speed/100;
      VCU_STATUS_OUT_FRAME.data.byte[4] = (byte)mg2rpmby100;

      //Send VCU Status
      Can0.sendFrame(VCU_STATUS_OUT_FRAME);

      //For testing only
      ++MSG_DATA;
  }
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
 * Simple printf for writing to an Arduino serial port.  Allows specifying Serial..Serial3.
 * 
 * Note (AK): This greatly affects timing of Metro lib and millis() 
 * 
 * const HardwareSerial&, the serial port to use (Serial..Serial3)
 * const Serial_&, the serial port to use (SerialUSB)
 * const char* fmt, the formatting string followed by the data to be formatted
 * 
 * int d = 65;
 * float f = 123.4567;
 * char* str = "Hello";
 * serial_printf(Serial, "<fmt>", d);
 * 
 * Example:
 *   serial_printf(Serial, "Sensor %d is %o and reads %1f\n", d, d, f) will
 *   output "Sensor 65 is on and reads 123.5" to the serial port.
 * 
 * Formatting strings <fmt>
 * %B    - binary (d = 0b1000001)
 * %b    - binary (d = 1000001)  
 * %c    - character (s = H)
 * %d/%i - integer (d = 65)\
 * %f    - float (f = 123.45)
 * %3f   - float (f = 123.346) three decimal places specified by %3.
 * %o    - boolean on/off (d = On)
 * %s    - char* string (s = Hello)
 * %X    - hexidecimal (d = 0x41)
 * %x    - hexidecimal (d = 41)
 * %%    - escaped percent ("%")
 *
 * void serial_printf(HardwareSerial& serial, const char* fmt, ...)   // to use (Serial..Serial3)
 * void serial_printf(Serial_& serial, const char* fmt, ...)          // to use (SerialUSB)
 */
void serial_printf(Serial_& serial, const char* fmt, ...) { 
    va_list argv;
    va_start(argv, fmt);

    for (int i = 0; fmt[i] != '\0'; i++) {
        if (fmt[i] == '%') {
            // Look for specification of number of decimal places
            int places = 2;
            if (fmt[i+1] >= '0' && fmt[i+1] <= '9') {
                places = fmt[i+1] - '0';
                i++;
            }
            
            switch (fmt[++i]) {
                case 'B':
                    serial.print("0b"); // Fall through intended
                case 'b':
                    serial.print(va_arg(argv, int), BIN);
                    break;
                case 'c': 
                    serial.print((char) va_arg(argv, int));
                    break;
                case 'd': 
                case 'i':
                    serial.print(va_arg(argv, int), DEC);
                    break;
                case 'f': 
                    serial.print(va_arg(argv, double), places);
                    break;
                case 'l': 
                    serial.print(va_arg(argv, long), DEC);
                    break;
                case 'h':
                    serial.print(va_arg(argv, int) == 0 ? "Low" : "High");
                    break;
                case 'o':
                    serial.print(va_arg(argv, int) == 0 ? "Off" : "On");
                    break;
                case 'y': //added this case for "Yes/No" :-)
                    serial.print(va_arg(argv, int) == 0 ? "No" : "Yes");
                    break;
                case 's': 
                    serial.print(va_arg(argv, const char*));
                    break;
                case 'X':
                    serial.print("0x"); // Fall through intended
                case 'x':
                    serial.print(va_arg(argv, int), HEX);
                    break;
                case '%': 
                    serial.print(fmt[i]);
                    break;
                default:
                    serial.print("?");
                    break;
            }
        } else {
            serial.print(fmt[i]);
        }
    }
    va_end(argv);
}
 //********************USB SERIAL INPUT FROM KEYBOARD *******I*********
 /* These routines use an automatic interrupt generated by SERIALEVENT to process keyboard input any time a string terminated with carriage return is received 
  *  from an ASCII terminal program via the USB port.  The normal program execution is interrupted and the checkforinput routine is used to examine the incoming
  *  string and act on any recognized text.
  * 
  */
   
void serialEventRun(void) 
{
   if (SerialDEBUG.available())checkforinput(); //If serial event interrupt is on USB port, go check for keyboard input           
}

void checkforinput()
{
  //Checks for input from Serial Port indicating configuration items most likely.  Handle each case.
  
   if (SerialDEBUG.available()) 
   {
    String input = Serial.readString();
    input.trim();                                          //remove whitespace (" ","\t","\v","\f","\r","\n")
    input.toCharArray(charBuf, 10);                        //convert string to char*
    
    String param = input.substring(0,1);                   //parse out parameter name (single character)
    String strValue = input.substring(1);                  //parse out parameter value
    int intValue = input.substring(1).toInt();             //convert value to numeric
    int inByte = param.charAt(0); //convert to char

    //serial_printf(Serial, "Input Value [%s]\n", charBuf);  //does not work with strings
    //Serial.print("Param [");Serial.print(param);Serial.print("]; Value [");Serial.print(strValue);Serial.print("]; intValue [");Serial.print(intValue);Serial.println("]");
    //serial_printf(Serial, "Param [%s]; strValue [%s]; intValue [%i]\n", param, charBuf, intValue);  //needs more work with strings :-(
    //Serial.println(inByte);

    switch (inByte)
     {
       case 'h':
       case 'm':
       case '?':
          printMenu();
          break;  
       case 'c': 
          resetSensorCounters();
          break;
       case 'd':
          toggleDebugLevel();
          break;
       case 'p':
          setTorqePerc(intValue);
          break;
       case 'g':
          setGearRatio(intValue);
          break;
       case 's':
          setSportMode(intValue);
          break;
       case 'b':
          setBrakePressure(intValue);
          break;
       case 'e':
          setEngineBraking(intValue);
          break;
       case 'r':
          setRampDuration(intValue);
          break;
       default:
          break;
       }
     /*
     String input = SerialDEBUG.readString();
     //SerialDEBUG.print("Input Value: ");SerialDEBUG.println(input);
     int number = input.toInt();
     //SerialDEBUG.print("Numeric Value: ");SerialDEBUG.println(number);
     inputValue = number;
     */
    }
}

short limitRange(short value, short min, short max)
{
  if (value > max)
  {
    return max;
  }
  else if (value < min)
  {
    return min;
  }
  else
  {
    return value;
  }
}

void setRampDuration(short value)
{
  RampDuration = limitRange(value, 0, 10000);
  ResetRgenRamp(false);
}

void setEngineBraking(short value)
{
  EngineBraking = limitRange(value, 0, 10);
}

void setBrakePressure(short value)
{
  BrakePressure = limitRange(value, 0, 10);
}

void setSportMode(int value)
{
  SportMode = (bool)value;
}

void setGearRatio(int value)
{
  RequestedGear = (bool)value;
}

void setTorqePerc(int value)
{
  int percVal = 0;
  
  if (value < 0)
  {
    percVal= 0;
  }
  else if (value > 100)
  {
    percVal= 100;
  }
  else
  {
    percVal= value;
  }
  
  TorquePercent = percVal;
}

void printMenu()
{
  SerialDEBUG.print("************ Jaguar VCU Commands ************\n\n");
  SerialDEBUG.print("  ? - Print this menu\n");
  SerialDEBUG.print("  c - Reset ISA Shunt counters\n");
  SerialDEBUG.print("  p - Perentage of usable Torque (0-100)\n");
  SerialDEBUG.print("  g - Select speed (1->Hi, 0->Lo)\n");
  SerialDEBUG.print("  s - Select SportMode (1->Sport, 0->Normal)\n");
  SerialDEBUG.print("  d - Toggle Debug Mode\n");
  SerialDEBUG.print("  b - Simulated Brake Pressure\n");
  SerialDEBUG.print("*********************************************\n\n");
}

debugLevels toggleDebugLevel()
{
  SerialDEBUG.print("Debug Mode: ");
  switch (_debug)
  {
    case NONE:
      SerialDEBUG.println("DETAIL");
      _debug = DETAIL;
      break;
    case DETAIL:
      SerialDEBUG.println("CSV");
      _debug = CSV;
      break;
    case CSV:
      SerialDEBUG.println("NONE");
      _csv_log_started = false;
      _debug = NONE;
      break;
  }
}

void resetSensorCounters()
{
  SerialDEBUG.print("Resetting ISA counters...");
  Sensor.RESTART(); // Reset Ah and kWh counters
  SerialDEBUG.print("Done!");
}

byte get_oilPWM()
{
  byte oilPWM = 0;
  if (gear != PARK)
  {
    oilPWM = map(TorqueReqest, 0, 1500, 40, 100);
    if (oilPWM < 40) oilPWM = 40;
    if (oilPWM > 120) oilPWM = 120;
  }
  if (SportMode) oilPWM = 60;
  return oilPWM;
}

void restoreTorque()
{
  if (TorqueRestoreOK && TorquePercent < 100)
  {
    TorquePercent++;
  }
}

void sportGears()
{
  if (SportMode)
  {
    if (mg2_speed < 2000 && CurrentGear) RequestedGear = LOW;
    else if (mg2_speed > 7000 && !CurrentGear) RequestedGear = HIGH;
  }
}

void loop() 
{
  can_status = true;
  Key_ON = get_keyOn(); // detect "Key ON" state
  Charge_Mode = get_chargeMode(); //detect "Charge Mode"
  vehicleMode = get_VehicleMode();
  
  if (Key_ON && !Charge_Mode)             // Driving Mode
  {
    analogWrite(OilPumpPWM, oil_power);   // set oil pump speed
    digitalWrite(InvPower, HIGH);         // turn on inverter, oil pump
  }
  else //if (!Key_ON && Charge_Mode)      // Charging/Programming Mode
  {
    analogWrite(OilPumpPWM, 0);           // set oil pump speed
    digitalWrite(InvPower, LOW);          // turn off inverter, oil pump
  }
  
  if (HV_Enabled)                         // Control Inverter only if HV is present
  {
    control_inverter();
  }
 
  Frames4MS();
  Frames100MS();
  Frames200MS();

  if (timer_1ms.check())
  {
    restoreTorque();
  }

  if (timer_100ms.check())
  {
    _millis = millis();
    //if (vehicleMode > PROGRAMMING)
    //{
      Control_HV(); //do not call this in DEBUG/PROG mode since 10ms delay is messing up loop timings
    //}
    sportGears();
    setGear();
  }
  
  if (timer_250ms.check())
  {
    oil_power = get_oilPWM();
  }

  if (timer_diag.check())
  {
    if (vehicleMode == 2 && Sensor.kwh < 0) resetSensorCounters(); //reset counters since battery was charged
    //TODO: Ideally this should reflect SOC vs kWh used
    VCU_STATUS_OUT_FRAME.data.byte[3] = Sensor.kwh/100; //set kWh used since full charge

    ///////////////////////////////////////////////////////////////
    //TODO: Coolant Temp
    ///////////////////////////////////////////////////////////////
    //int temp = map(analog1, 0, 765, 80, 125);
    //int temp = map(analog1, 0, 765, 150, 165);
    //SerialDEBUG.print("Temp: ");SerialDEBUG.println(temp);
    ///////////////////////////////////////////////////////////////

    switch (_debug)
    {
      case DETAIL:
        diag_mth();
        break;
      case CSV:
        csv_info();
        break;
      default:
        SerialDEBUG.print("Millis: "); SerialDEBUG.println(_millis);
        //SerialDEBUG.print(CalcTorque( 0, BrakePressure, 3000));SerialDEBUG.print(",");SerialDEBUG.print(RegenPercentage);SerialDEBUG.print(",");SerialDEBUG.print(RampStartTime); SerialDEBUG.println("");
        break;
    }

    handle_wifi();
    
    //SerialDEBUG.println(TorqueReqest);
    //SerialDEBUG.println(gear);
    
    //digitalWrite(13,!digitalRead(13));
  }
}
